from typing import Union, List, Mapping, Sequence

from ruko.binary_socket import BinarySocket
from ruko.command import Command
from ruko.empty_type import Empty
from ruko.empty_type import EmptyType
from ruko.serialize import with_len, json_to_bytes, bytes_to_json, DType
from ruko.utils import recursive_merge

Index = Union[str, list, tuple, int, None]
Json = Union[dict, list, int, float, bool, None]


class RukoClient:
    """Raw client to execute commands to server"""

    def __init__(self, ip: str, port: int):
        self.socket = BinarySocket(ip, port)

    def close(self):
        self.socket.close()

    @staticmethod
    def parse_key(key: Index) -> list:
        if key is ...:
            key = '---'
        if isinstance(key, str):
            return [key]
        if isinstance(key, int):
            return [str(key)]
        if key is None or key is ():
            return []
        return list(map(str, key))

    def get(self, key: Index = None, fields: List[str] = None, exclude: List[str] = None) -> Union[Json, EmptyType]:
        # TODO: Support dotted filters in fields and exclude
        key = self.parse_key(key)
        fields = fields or []
        exclude = exclude or []
        d = self._send(
            bytes([Command.GET]) +
            self._encode_key(key) +
            with_len(b''.join(with_len(i.encode()) for i in fields), len(fields)) +
            with_len(b''.join(with_len(i.encode()) for i in exclude), len(exclude))
        )
        return d

    def set(self, key: Index, value: Json, fallback=None):
        key = self.parse_key(key)
        return self._send(
            bytes([Command.SET]) +
            self._encode_key(key) +
            json_to_bytes(value, fallback)
        )

    def delete(self, key: Index):
        return self._send(
            bytes([Command.DELETE]) +
            self._encode_key(key)
        )

    def declare(self, key: Index, dtype: DType, indices: List[str] = None):
        key = self.parse_key(key)
        data = (
                bytes([Command.DECLARE]) +
                self._encode_key(key) +
                bytes([dtype])
        )
        data += with_len(b''.join(with_len(i.encode()) for i in indices or []), length=len(indices))
        return self._send(data)

    def lput(self, key: Index, value: Json, fallback=None):  # TODO: Detect error (append on dict)
        key = self.parse_key(key)
        return self._send(
            bytes([Command.LPUT]) +
            self._encode_key(key) +
            json_to_bytes(value, fallback)
        )

    def create_mapping(self, key: Index, pattern: Index):
        key = self.parse_key(key)
        pattern = self.parse_key(pattern)
        return self._send(
            bytes([Command.CREATE_MAPPING]) +
            self._encode_key(key) +
            self._encode_key(pattern)
        )

    def delete_mapping(self, key: Index):
        key = self.parse_key(key)
        return self._send(
            bytes([Command.DELETE_MAPPING]) +
            self._encode_key(key)
        )

    def get_mappings(self):
        return [(name, pattern.split('\x19')) for name, pattern in self._send(bytes([Command.GET_MAPPINGS]))]

    def linsert(self, key: Index, index: int, value: Json):
        # TODO: Add linsert command
        data = self.get(key) or []
        data.insert(index, value)
        self.set(key, data)

    def dupdate(self, key: Index, element=None, **params):
        # TODO: Add dupdate command
        # TODO: Detect error (update on list)
        key = self.parse_key(key)
        val = self.get(key) or {}
        val.update(element, **params)
        self.set(key, val)

    def dmerge(self, key: Index, element=None, **params):
        # TODO: Add dmerge command
        # TODO: Detect error (merge on list)
        key = self.parse_key(key)
        new_dict = dict(element, **params)
        val = self.get(key) or {}
        val = dict(recursive_merge(val, new_dict))
        self.set(key, val)

    def dgetkeys(self, key: Index):
        # TODO: Add dgetkeys command
        key = self.parse_key(key)
        val = self.get(key)
        if isinstance(val, Mapping):
            return list(val.keys())
        elif isinstance(val, Sequence):
            return list(range(len(val)))
        return Empty

    def contains(self, key: Index, item: str) -> bool:
        # TODO: Add dcontains command
        return item in self.dgetkeys(key) or []

    def dgetvalues(self, key: Index):
        # TODO: Add dgetkeys command
        key = self.parse_key(key)
        return list((self.get(key) or {}).values())

    def clear(self, key: Index):
        # TODO: Add clear command
        self.set(key, type(self.get(key))())

    def _encode_key(self, key):
        return with_len(b''.join(with_len(i.encode()) for i in key), length=len(key))

    def _send(self, data: bytes) -> Union[Json, EmptyType]:
        ret = self.socket.send(data)

        if not ret:
            return Empty
        return bytes_to_json(ret)[0]



def recursive_merge(a, b):
    """
    Returns generator for merged dict of a and b
    Usage:
        >>> dict(recursive_merge({'a': {'b': 2}}, {'a': {'c': 3}}))
        {'a': {'c': 3, 'b': 2}}
    """
    for k in set(a.keys()) | set(b.keys()):
        if k in a and k in b:
            if isinstance(a[k], dict) and isinstance(b[k], dict):
                yield k, dict(recursive_merge(a[k], b[k]))
            else:
                yield k, b[k]
        elif k in a:
            yield k, a[k]
        else:
            yield k, b[k]



import struct
from enum import IntEnum
from typing import Any, Callable


class DType(IntEnum):
    EMPTY = 0
    NULL = 1
    BOOL = 2
    INT = 3
    FLOAT = 4
    STRING = 5
    LIST = 6
    DICT = 7


def bytes_to_u32(x):
    return struct.unpack('<L', x)[0]


def u32_to_bytes(x) -> bytes:
    return bytes(struct.pack('<L', x))


def bytes_to_f32(x):
    return struct.unpack('<f', x)[0]


def f32_to_bytes(x) -> bytes:
    return bytes(struct.pack('<f', x))


def bytes_to_i32(x):
    return struct.unpack('<i', x)[0]


def i32_to_bytes(x):
    return struct.pack('<i', x)


def with_len(bs: bytes, length=None) -> bytes:
    if length is None:
        length = len(bs)
    return u32_to_bytes(length) + bs


def bytes_to_null(_x, p=0) -> (None, int):
    return None, p


def bytes_to_bool(x, p=0) -> (bool, int):
    return x[0] != 0, p + 1


def bytes_to_int(x, p=0) -> (int, int):
    n = bytes_to_i32(x[p:p + 4])
    return n, p + 4


def bytes_to_float(x, p=0) -> (int, int):
    n = bytes_to_f32(x[p:p + 4])
    return n, p + 4


def bytes_to_str(x, p=0) -> (str, int):
    sz = bytes_to_u32(x[p:p + 4])
    p += 4
    data = x[p:p + sz].decode()
    p += sz
    return data, p


def bytes_to_list(x, p=0) -> (list, int):
    ls = []
    sz = bytes_to_u32(x[p:p + 4])
    p += 4
    for _ in range(sz):
        obj, p = bytes_to_json(x, p)
        ls.append(obj)
    return ls, p


def bytes_to_dict(x: bytes, p=0) -> (dict, int):
    d = {}
    sz = bytes_to_u32(x[p:p + 4])
    p += 4
    for _ in range(sz):
        key, p = bytes_to_str(x, p)
        d[key], p = bytes_to_json(x, p)

    return d, p


def bytes_to_json(x: bytes, p=0) -> (Any, int):
    func = {
        DType.NULL: bytes_to_null,
        DType.BOOL: bytes_to_bool,
        DType.INT: bytes_to_int,
        DType.FLOAT: bytes_to_float,
        DType.STRING: bytes_to_str,
        DType.LIST: bytes_to_list,
        DType.DICT: bytes_to_dict
    }[DType(x[p])]
    a, b = func(x, p + 1)
    return a, b


def raise_key_error(x, _key):
    raise ValueError('Cannot serialize: {}'.format(x))


def json_to_bytes(a: Any, fallback=None, key=()) -> bytes:
    func = {
        type(None): lambda x: bytes([DType.NULL]),
        bool: lambda x: bytes([DType.BOOL, int(x)]),
        int: lambda x: bytes([DType.INT]) + u32_to_bytes(x),
        float: lambda x: bytes([DType.FLOAT]) + f32_to_bytes(x),
        str: lambda x: bytes([DType.STRING]) + with_len(x.encode()),
        list: lambda x: bytes([DType.LIST]) + with_len(b''.join(
            json_to_bytes(v, fallback, key + (i,)) for i, v in enumerate(x)
        ), len(x)),
        dict: lambda x: bytes([DType.DICT]) + with_len(b''.join(
            with_len(k.encode()) + json_to_bytes(v, fallback, key + (k,)) for k, v in x.items()
        ), len(x))
    }.get(type(a))  # type: Callable
    if not func:
        return (fallback or raise_key_error)(a, key)
    return func(a)



from functools import wraps

from ruko.ruko_client import RukoClient


class RContext:
    """
    Class to handle database connection contexts. Using
    explicit contexts is not mandatory

    Usage:
        with db.context:
            db['notes'].append({})

        @db.context
        def get_notes():
            return db['notes'].get()
    """

    def __init__(self, *params, obj=None):
        self.params = params
        self.obj = obj or self

    def create_conn(self):
        return RukoClient(*self.params)

    def get(self):
        if not hasattr(self.obj, 'rkconn'):
            self.obj.rkconn = self.create_conn()
        return self.obj.rkconn

    def __call__(self, *args, **kwargs):  # As decorator
        if kwargs or len(args) != 1 or not callable(args[0]):
            raise ValueError('Improper usage of context. Check docstring')

        func = args[0]

        @wraps(func)
        def wrapper(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return wrapper

    def __enter__(self):
        if not hasattr(self.obj, 'rkconn'):
            self.obj.rkconn = self.create_conn()

    def close(self):
        if hasattr(self.obj, 'rkconn'):
            self.obj.rkconn.close()
            del self.obj.rkconn

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



import socket
from contextlib import suppress
from threading import Lock

from ruko.serialize import bytes_to_u32, u32_to_bytes


class BinarySocket:
    """A client for simple length-prefixed socket communication"""

    def __init__(self, ip, port):
        self.ip = ip
        self.port = port
        self.socket = self.open_socket()
        self.lock = Lock()

    def open_socket(self):
        s = socket.socket()
        s.connect((self.ip, self.port))
        return s

    def send(self, msg: bytes) -> bytes:
        try:
            return self._send(msg)
        except (ConnectionAbortedError, OSError):
            print('Reconnecting...')
            self.socket = self.open_socket()
            return self._send(msg)

    def _send(self, msg: bytes):
        len_bytes = u32_to_bytes(len(msg))
        bs = len_bytes + msg

        with self.lock:
            self.socket.sendall(bs)
            length = bytes_to_u32(self._read_bytes(4))
            return self._read_bytes(length)

    def close(self):
        with self.lock:
            self._close()

    def _close(self):
        with suppress(OSError):
            self.socket.shutdown(2)
        self.socket.close()

    def _read_bytes(self, n: int) -> bytes:
        buf = b''
        while len(buf) < n:
            new_data = self.socket.recv(n - len(buf))
            if not new_data:
                self._close()
                raise ConnectionAbortedError
            buf += new_data
        return buf



class EmptyType(object):
    """A sentinel value when nothing is returned from the database"""

    def __new__(cls):
        return Empty

    def __reduce__(self):
        return EmptyType, ()

    def __bool__(self):
        return False

    def __repr__(self):
        return 'Empty'


Empty = object.__new__(EmptyType)



from setuptools import setup

setup(
    name='ruko',
    version='0.1.1',
    license='MIT',
    description='A dead simple high performance JSON database',
    author='Matthew Scholefield',
    author_email='matthew331199@gmail.com',
    url='https://github.com/rukodb/ruko',
    packages=['ruko'],
    install_requires=[],
    keywords='ruko database in-memory',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',

        'Topic :: Database :: Database Engines/Servers',
        'Topic :: Software Development :: Libraries :: Python Modules',

        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ]
)



from collections import MutableMapping

from typing import Iterator, MutableSequence, Union

from ruko.context import RContext
from ruko.empty_type import Empty
from ruko.ruko_client import RukoClient, Index, Json


class RDict(MutableMapping, MutableSequence):
    """High level Python interface to Ruko database"""

    @classmethod
    def client(cls, ip: str = "127.0.0.1", port: int = 44544, key_error=KeyError):
        """Create a new connection"""
        return cls([], RContext(ip, port), key_error)

    def __init__(self, db_key, context: RContext, key_error=KeyError):
        """Typically, you want to use RDict.client() instead"""
        self.key = RukoClient.parse_key(db_key)
        self.context = context
        self.key_error = key_error

    @property
    def rk(self):
        return self.context.get()

    def __iter__(self) -> Iterator[str]:
        keys = self.rk.dgetkeys(self.key)
        if keys is Empty:
            return iter(self.get())
        return iter(keys)

    def items(self) -> Iterator[tuple]:
        return self().items()

    def insert(self, index: int, obj: Json) -> None:
        self.rk.linsert(self.key, index, obj)

    def keys(self):
        return self.rk.dgetkeys(self.key)

    def values(self):
        return self.rk.dgetvalues(self.key)

    def by(self, key: str) -> 'RDict':
        return self.child(self.sub_key([key + ':']))

    def get(self, item=None, default=None, fields=None, exclude=None):
        item = RukoClient.parse_key(item)
        val = self.rk.get(self.sub_key(item), fields, exclude)
        if val is Empty:
            return default
        return val

    def update(self, element=None, **params):
        return self.rk.dupdate(self.key, element, **params)

    def merge(self, element=None, **params):
        return self.rk.dmerge(self.key, element, **params)

    def clear(self):
        self.rk.clear(self.key)

    def append(self, value):
        self.rk.lput(self.key, value)

    @staticmethod
    def concat(a: list, b: list):
        if a and a[-1].endswith(':') and b:
            a = list(a)
            a[-1] = a[-1] + (b[0] or '~')
            return a + b[1:]
        return a + b

    def sub_key(self, b: list):
        return self.concat(self.key, b)

    def child(self, full_key) -> 'RDict':
        return RDict(full_key, self.context, self.key_error)

    def setdefault(self, key, default=None) -> 'RDict':
        key = RukoClient.parse_key(key)
        new_key = self.concat(self.key, key)
        cur_val = self.rk.get(new_key)
        if cur_val is Empty:
            self.rk.set(new_key, default)
        return self.child(new_key)

    def create_mapping(self, key, filters) -> 'RDict':
        key = RukoClient.parse_key(key)
        new_key = self.sub_key(key)
        if isinstance(filters, list):
            if filters and isinstance(filters[0], (str, int)):
                if any(isinstance(i, str) and '.' in i for i in filters):
                    filters = [i.split('.') for i in filters]
                else:
                    filters = [filters]
            else:
                pass  # Already in right format
        else:
            if isinstance(filters, str) and '.' in filters:
                filters = [filters.split('.')]
            else:
                filters = [RukoClient.parse_key(filters)]
        for filt in filters:
            self.rk.create_mapping(new_key, filt)
        return self.child(new_key)

    def delete_mapping(self, key):
        key = RukoClient.parse_key(key)
        self.rk.delete_mapping(self.sub_key(key))

    def get_mappings(self):
        return self.rk.get_mappings()

    def __contains__(self, item):
        return self.rk.contains(self.key, item)

    def __getitem__(self, item: Index) -> Union['RDict', str, int, list, dict, bool, None]:
        item = RukoClient.parse_key(item)
        return self.child(self.sub_key(item))

    def __call__(self, fields=None, exclude=None):
        val = self.rk.get(self.key, fields, exclude)
        if val is Empty:
            raise self.key_error(self.key[-1])
        return val

    def __delitem__(self, key):
        key = RukoClient.parse_key(key)
        return self.rk.delete(self.sub_key(key))

    def __setitem__(self, key, value):
        key = RukoClient.parse_key(key)
        self.rk.set(self.sub_key(key), value)

    def __len__(self):  # TODO: Add len command
        return len(self.get())

    def __repr__(self):
        return '<RDict key={}>'.format('.'.join(map(str, self.key)))

    def __eq__(self, other):
        if isinstance(other, RDict):
            other_val = other.get()
        else:
            other_val = other
        my_val = self.get()
        return my_val == other_val
